#!/usr/bin/perl

use strict;
use LWP::UserAgent;
use Digest::MD5 qw(md5_base64);

my ($protocol,$base_path) = ('https://','/cgi/');

use constant EXIT_OK => 0;
use constant EXIT_CONNECTION_ERROR => 1;
use constant EXIT_WRONG_FLAG => 5;
use constant EXIT_MISSING_FUNCTIONALITY => 9;
use constant EXIT_RESPONSE_TIMEOUT => 13;   # usually set by GS
use constant EXIT_STATUS_UNKNOWN => 17;     # unneeded
use constant EXIT_ERROR => 21;
use constant EXIT_PROTOCOL_VIOLATION => 25;

my $search_check_length = 8;

#---------------------------------- start off with parameters
#{{{
sub msg;

unless(@ARGV >= 4)  {
  msg('internal problem',"$0 called with only ".scalar(@ARGV)." arguments (which is less than 4!)");
  exit(EXIT_ERROR);
}
#}}}
my ($action,$ip,$flag_id,$flag) = @ARGV;
#{{{
sub get_env {
  my ($name) = @_;
  return undef unless(defined $ENV{$name});
  return 0 unless($ENV{$name} =~ /(\d+)/);
  return $1;
}
#}}}
my ($team_num,$flag_num,$round_num,$timeout) = 
    (get_env('CTFGAME_TEAMNUM'),get_env('CTFGAME_FLAGNUM'),get_env('CTFGAME_ROUNDNUM'),get_env('CTFGAME_TIMEOUT'));

$timeout *= 0.9 if(defined $timeout);

#----------------------------------  tools

sub msg {
  my ($public,$secret) = @_;
#{{{
  if (defined $public) {
    chomp($public);
    print $public,"\n";
    $secret=$public unless(defined $secret);
  }
  chomp($secret);
  print STDERR 'S> ',$secret,"\n";
}
#}}}

my $ua = LWP::UserAgent->new;
$ua->timeout($timeout) if(defined $timeout);
$ua->agent('');
$ua->credentials("$ip:443",'IT Sec',$ENV{'TEST_USER'},$ENV{'TEST_PASS'}) if(defined $ENV{'TEST_USER'});

sub myget {
  my ($url) = @_;
  my $response = $ua->get($protocol.$ip.$base_path.$url);
  if ($response->is_success) {
    msg(undef,"called $url");
    return $response->decoded_content;
  } else {
    msg("call failed with ".$response->status_line,
        "call failed with ".$response->status_line."\nURL=$url");
  }
}

sub urlencode {
  my ($str) = @_;
  $str =~ s/([^A-Za-z0-9])/sprintf("%%%02X", ord($1))/seg;
  return $str;
}

#---------------------------------- 

my ($caption,$content,$shout);

sub generate_data {
  # TODO: to be replaced with more funny things/decoys/stuff
  $caption = $flag_id;
  $content = $flag;

  $shout = md5_base64($content);
}

#---------------------------------- 

sub place_flag {
  # fetch first page
  my $html = myget('index.bas');
  unless($html =~ /shoutbox.*shouts/si) {
    msg("shoutbox missing");
    return EXIT_MISSING_FUNCTIONALITY;
  }
  # go to upload page
  $html = myget('index.bas?inc=upload.bas');
  unless($html =~ /upload of some data/si) {
    msg("upload form missing");
    return EXIT_MISSING_FUNCTIONALITY;
  }
  # upload form
  $html = myget('index.bas?inc=upload.bas&caption='.urlencode($caption).'&content='.urlencode($content));
  unless($html =~ /Store successful/si) {
    msg("upload failed");
    return EXIT_MISSING_FUNCTIONALITY;
  }
  # finaly finished
  return EXIT_OK;
}

sub check_search {
  my ($deep) = (rand()<0.5 ? 1 : 0);
  msg(undef,"decided to test searching capabilities (deep = $deep)");
  # get search form
  my $html = myget('index.bas?inc=search.bas');
  unless($html =~ /Search for something/si) {
    msg("missing search-form");
    return EXIT_MISSING_FUNCTIONALITY;
  }
  # decide upon search term
  my $term;
  if ($deep) {
    $term = substr($content,rand(length($content)-$search_check_length),$search_check_length);
  } else {
    $term = substr($caption,rand(length($caption)-$search_check_length),$search_check_length);
  };
  msg(undef,"search for '$term'");
  # call search
  $html = myget('index.bas?inc=search.bas&term='.urlencode($term).($deep?'&deep=1':''));
  if(index($html,$caption)<0) {
    msg("search did not reveal correct filename");
    return EXIT_MISSING_FUNCTIONALITY;
  }
  return EXIT_OK;
}

sub check_shoutbox {
  msg(undef,"decided to test shoutbox capabilities");
  # fetch first page
  my $html = myget('index.bas');
  unless($html =~ /shoutbox.*shouts/si) {
    msg("shoutbox missing");
    return EXIT_MISSING_FUNCTIONALITY;
  }
  # post shout
  $html = myget('index.bas?action=nusb&shout='.urlencode($shout));
  if(index($html,$shout)<0) {
    msg("shoutbox mal function");
    return EXIT_MISSING_FUNCTIONALITY;
  }
  return EXIT_OK;
}

sub check_flag {
  # fetch first page
  my $html = myget('index.bas');
  unless($html =~ /shoutbox.*shouts/si) {
    msg("shoutbox missing");
    return EXIT_MISSING_FUNCTIONALITY;
  }
  # check some random functionality
  if (rand()<0.5) {
    my $ret = check_search();
    return $ret unless($ret);
  }
  if (rand()<0.5) {
    my $ret = check_shoutbox();
    return $ret unless($ret);
  }
  # go to download page
  $html = myget('index.bas?inc=download.bas');
  unless($html =~ /download data/si) {
    msg("list of download documents missing");
    return EXIT_MISSING_FUNCTIONALITY;
  }
  if(index($html,$caption)<0) {
    msg("item which is to be requested is missing");
    return EXIT_MISSING_FUNCTIONALITY;
  }
  # download document
  $html = myget('index.bas?inc=download.bas&gimme='.urlencode($caption));
  if(index(lc($html),"content of $caption")<0) {
    msg("wrong header");
    return EXIT_MISSING_FUNCTIONALITY;
  }
  if (index(lc($html),lc($flag))<0) {
    msg("wrong flag");
    return EXIT_WRONG_FLAG;
  }
  # finally finished
  return EXIT_OK;
}

#----------------------------------  main
#{{{
if ($action eq 'store') {
  generate_data();
  exit(place_flag());
} elsif($action eq 'retrieve') {
  generate_data();
  exit(check_flag());
} elsif($action eq 'test') {
  generate_data();
  exit(check_search() || check_shoutbox());
} else {
  msg('internal problem',"$0 called with '$action' as 1st argument (should be store or retrieve!)");
  exit(EXIT_ERROR);
}
#}}}
# vim: et tabstop=2 foldmethod=marker foldenable foldlevel=0 lbr ai nu fdc=1
