#!/usr/bin/perl 

use strict;
use warnings;

use LWP::Simple;

my $APP_PORT = 7000;
my $RAND_STRING_LENGTH=30;

my $task = $ARGV[0];

if ($task eq 'test') {
    print STDERR "Everything is fine.";
    print "Everything is fine.";
    exit 0;
}

if ($task eq 'store') {
    my $ip     = $ARGV[1];
    my $flagid = $ARGV[2];
    my $flag   = $ARGV[3];

    my $result = get("gopher://$ip:$APP_PORT/7FLAGSTORE$flagid?$flag");
    if (! defined $result) {
        print STDERR "Failed fetching via gopher";
        print "Failed fetching via gopher";
        exit 1;
    }
    if ($result =~ /FLAGZ HAS BIN STORD/) {
        print STDERR "Stored flag successfully";
        print "Everything is fine.";
        exit 0;
    }
    else {
        print STDERR "Failed storing flag: $result";
        print "Failed storing flag";
        exit 21;
    }
}

if ($task eq 'retrieve') {
    my $ip     = $ARGV[1];
    my $flagid = $ARGV[2];
    my $flag   = $ARGV[3];

    # test palindrome functionality
    my $string;
    my $is_palindrome;
    if (rand() < 0.5) {
        $string = get_random_palindrome();
        $is_palindrome = 1;
    }
    else {
        $string = get_random_string();
        $is_palindrome = 0;
    }
    my $palindromecheck = get("gopher://$ip:$APP_PORT/7IZPALEENDROAM?$string");
    if (! defined $palindromecheck) {
        print STDERR "Failed fetching (IZPALEENDROAM) via gopher";
        print "Failed fetching via gopher";
        exit 1;
    }
    if (! (($palindromecheck =~ /KAY/ && $is_palindrome) ||
           ($palindromecheck =~ /OWAI/ && ! $is_palindrome))) {
        # incorrect answer
        print STDERR "Palindrome checking failed: $palindromecheck";
        print "Palindrome checking failed";
        exit 9;
    }

    my $result = get("gopher://$ip:$APP_PORT/0$flagid");
    if (! defined $result) {
        print STDERR "Failed fetching via gopher";
        print "Failed fetching via gopher";
        exit 1;
    }
    if ($result =~ /$flag/) {
        print STDERR "Successfully retrieved flag.";
        print "Successfully retrieved flag";
        exit 0;
    }
    else {
        print STDERR "Wrong/no flag retrieved: " . $result;
        print "Wrong/no flag retrieved";
        exit 5;
    }
}

sub get_random_palindrome {
    my $str = get_random_string($RAND_STRING_LENGTH / 2);
    return $str . reverse($str);
}

sub get_random_string {
    my $length = shift || $RAND_STRING_LENGTH;
    my $string = "";
    for (my $i = 0; $i < $length; $i++) {
        $string .= chr(int(rand(60))+65);
    }
    return $string;
}
