import sys
import os
import socket
import re
import sha
from random import randint
from time import time
from sys import stdout, stderr


# Error codes
# Please make sure you have read http://ctf.hcesperer.org/gameserver/errorcodes.html
ERR_OK = 0 # All OK
ERR_CONNECTION = 1 # Connection refused / connection attempt timed out
ERR_WRONGFLAG = 5 # A wrong flag / no flag was returned
ERR_FUNCLACK = 9 # The service lacks functionality
ERR_TIMEOUT = 13 # Done automatically by the scorebot under normal circumstances
ERR_UNKNOWN = 17 # Temporary status -- do not use unless you've got a very good reason to do so
ERR_GENERIC = 21 # Be sure to include a descriptive message if using this error
ERR_PROTOCOL = 25 # Protocol violation

# Add error descriptions here.
# Note (to zeri): do not insult people. ;-)
shit = {
    'create': ("Creating webhosting user failed", ERR_FUNCLACK),
    'connect': ("Connection failed", ERR_CONNECTION),
    'curtime': ("Uploading /whatsthetime.html failed", ERR_FUNCLACK),
    'mkdir': ("mkdirs command failed", ERR_FUNCLACK),
    'upload': ("Upload failed", ERR_FUNCLACK),
    'download': ("Download failed", ERR_FUNCLACK),
    'get': ("GET function failed", ERR_FUNCLACK),
    'createacc': ("User routine failed: couldn't create account", ERR_GENERIC),
    'userlogin': ("User routine failed: couldn't log in", ERR_GENERIC),
    'flag': ("Wrong flag / no flag returned", ERR_WRONGFLAG),
    'allok': ("Everything is fine. How'd you do it?", ERR_OK),
    'illegalparams': ("Illegal command line parameters", ERR_GENERIC),
    }

def die(reason):
    """This function should be called to terminate the testscript,
whether successfully or unsuccessfully. An exitcode must always be
specified."""
    try: msg, exitcode = shit[reason]
    except: msg, exitcode = reason, 1
    stdout.write(msg)
    stdout.flush()
    exit(exitcode)

def randflag():
    """This function can be used to generate random (invalid, but
indistinguishable from real) flags."""
    flagpart = "ABCDEFabcdef0123456789"
    return ''.join([flagpart[randint(0, len(flagpart) - 1)] for i in range(8)])

debug = False

class LineReader:
    buf = ''     
    def __init__(self, socket):
        self.s = socket        
    def readline(self):        
        while True:            
            pos = self.buf.find("\n")
            if pos != -1:            
                line, self.buf = self.buf[:pos], self.buf[pos + 1:]
                return line                                        
            frag = self.s.recv(8192)                               
            if frag == None: raise EOFException()                  
            if len(frag) == 0: raise EOFException()                
            frag = frag.replace("\r", "")                          
            if len(frag) == 0:                                     
                buf, self.buf = self.buf, ''                       
                return buf # EOF                                   
            self.buf = self.buf + frag                             
    
class CommandExecutor:
    def __init__(self):
        self.redict = {}
    def RECompile(self, rtc):
        try: return self.redict[rtc]
        except:
            rx = re.compile(rtc)
            self.redict[rtc] = rx
            return rx
    def sendall(self, s, wts):
        if debug: sys.stderr.write(wts)
        s.sendall(wts)
    def DoCommand(self, sockaddr, command, params, expect, body=None):
        expect = [(lambda r: self.RECompile(r) if r else False)(r) for r in expect]
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        lr = LineReader(s)
        try: s.connect(sockaddr)
        except: die('connect')
        self.sendall(s, "%s\n" % " ".join([(lambda f: f if f.find(' ') == -1 else '"%s"' % f)(f) for f in command]))
        if len(params):
            sts = "\n".join(["\n".join(["%s: %s" % (key, item) for item in params[key]]) for key in params] + [''])
            self.sendall(s, sts)
        self.sendall(s, "\n")
        rewait = expect.pop(0)
        match1 = None
        while rewait:
            try: line = lr.readline()
            except: return None, None, None
            if debug: print line
            if len(line) == 0: break
            match1 = rewait.match(line)
            if match1: break
        if body:
            rewait = expect.pop()
            self.sendall(s, body)
            while rewait:
                try: line = lr.readline()
                except: return None, None, None
                if len(line) == 0: return None
                match2 = rewait.match(line)
                if match2: return match1.groups() if match1 else None, match2.groups() if match2 else None, lr
        else: return match1.groups() if match1 else None, None, lr
        return None, None, None
                      

def MakeSessionID(sesslen, salt):
    foo = sha.sha("%d%s" % (sesslen, salt)).hexdigest().encode('base64')
    return foo[:sesslen]

class Connection:
    def __init__(self, host, port, salt=None, user=None, password=None, fullname=None):
        self.session = None
        self.salt = salt if salt else time()
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.fullname = fullname
        self.ce = CommandExecutor()
    def SetSession(self, sid):
        res,f,b = self.ce.DoCommand((self.host, self.port), ['SESSION', sid], {}, ('^([0-9]{3}).*$',))
        return len(res) and (res[0] in ('200', '400'))
    def InitSession(self):
        if self.session: return True
        for sesstotry in [MakeSessionID(i, self.salt) for i in range(1, 15)]:
            if self.SetSession(sesstotry):
                self.session = sesstotry
                return True
        return False            
    def get(self, path, vhost=None, cookies=None, body=None):
        if not self.InitSession(): return None
        parms = {'host': (vhost,), 'session': (self.session,)} if vhost else {'session': (self.session,)}
        if cookies: parms['cookie'] = cookies
        res,f,r = self.ce.DoCommand((self.host, self.port),
                                    ['GET' if not body else 'POST', path, 'HTTP/1.1'],
                                    parms,
                                    (None,
                                     '^HTTP/1.1 ([0-9]{3}) .*',) if body else ('^HTTP/1.1 ([0-9]{3}) .*',None,),
                                    body)
        res = res if res else f
        if not res: return None
        if (len(res) != 1) or (res[0] != "200"): return None
        lines = []
        while len(r.readline().strip()): pass
        while True:
            try: line = r.readline()
            except: break
            # if len(line.strip()) == 0: break
            lines.append(line)
        return "\n".join(lines)
    def createuser(self):
        res,f,r = self.ce.DoCommand((self.host, self.port),
                                ['CREATE', self.user, self.fullname, self.password],
                                    {},
                                    ('^([0-9]{3})( .*)', None))
        return int(res[0]), res[1:]
    def download(self, path):
        res,f,r = self.ce.DoCommand((self.host, self.port),
                                    ['DTP', 'DOWNLOAD', self.user, self.password, path],
                                    {},
                                    ('([0-9]{3}) Thank you, "(.+?)", your file is ([0-9]+) (lines?) long in encoded form', None))
        if res and (len(res) >= 2) and (res[0] == '200') and (res[1] == self.fullname):
            lines = int(res[2])
            if lines > 1024: return None, None
            lr = "".join([r.readline() for i in range(lines)]).decode('base64')
            return lr
        return None
    def upload(self, path, fc):
        lines = fc.encode('base64')
        res,f,r = self.ce.DoCommand((self.host, self.port),
                                    ['DTP', 'UPLOAD', self.user, self.password, path],
                                    {'Lines': (str(len(lines.split("\n"))),)},
                                    ('^(200) .*', '^([0-9]{3}) Thank you, (.+?), .*'), lines + "\n")
        return (f != None) and (len(f) >= 2) and (f[0] == '200') and (f[1] == self.fullname)
    def mkdirs(self, path):
        res,f,r = self.ce.DoCommand((self.host, self.port),
                                    ['DTP', 'MKDIRS', self.user, self.password, path],
                                    {},
                                    ('^([0-9]+) .*', None))
        return (len(res) >= 1) and (res[0] == '200')

yesno = lambda: (int(str(time())[-1]) % 2) == 0

class TestFunctions:
    def store0(self, c, r, t, s, fid, f):
        if not c.mkdirs("/foo/%s/" % fid): die('mkdir')
        if not c.upload("/foo/%s/bar.html" % fid,
                        """<html><head><title>%s</title></head><body><p>
Flags! They belong to us. All!</p></body></html>""" % f): die('upload')
    def store1(self, c, r, t, s, fid, f):
        if not c.upload("/A%s.cgi" % fid, """#!/usr/bin/env python

import sys

encdata = "\\n".join([line for line in sys.stdin])
encdata = "<p>Verschl&uuml;sselte Daten: %%s</p>" %% encdata.encode('base64') if len(encdata.strip()) else ""

print \"\"\"Content-type: text/html

<html><head><title>64Corp TM -- BASE 64 Services since 1964</title></head>
<body>
<h1>64CORP WELCOME!!1!</h1>
<p><u>Herzlich Willkommen</u> zur 64CORPoration! Wir wandeln Ihre Zeichenketten
Sicher, zuverlaessig und vertrauenswuerdig um.</p>
%%s
<form method="POST" action="/%s.cgi">
<p>Zu verschl&uuml;sselnder Schadcode (ScriptKiddieTM Service): <br />
<textarea name="schadcode" cols="40" rows="40"></textarea><input type="submit" value="Daq yInej" /></p>
</form></body></html>
\"\"\" %% encdata

""" % fid): die('upload')
    def store2(self, c, r, t, s, fid, f):
        if not c.upload("/Alogin_%s.ipy" % fid, """def createuser(req, res, db, s, in, user, pass):
\tlog("Creating table users")
\tsucc = db.execute("CREATE CACHED TABLE users (uid INTEGER IDENTITY, u_name VARCHAR(256), u_pass VARCHAR(256))")
\tlog(succ)
\tsb = StringBuilder.new()
\tsb.append("INSERT INTO users (u_name, u_pass) values('")
\tsb.append(user)
\tsb.append("', '")
\tsb.append(pass)
\tsb.append("')")
\tlog(sb.toString())
\tsucc = db.execute(sb.toString())
\tlog(succ)
\tres.setContentType("text/plain")
\tsucc = db.execute("SHUTDOWN")
\tlog(succ)
\treturn "Jup, Account created"
def store(req, res, db, s, in, user, pass, key, value):
\tsb = StringBuilder.new()
\tsb.append("SELECT u_pass from users where u_name='")
\tsb.append(user)
\tsb.append("'")
\tlog(sb.toString())
\tres.setContentType("text/plain")
\tqres = db.query(sb.toString())
\tif isnull(qres):
\t\treturn "Error: login incorrect"
\tfor datensatz in qres:
\t\tpword = datensatz.getat(0)
\t\tif pass == pword:
\t\t\ts.put("authenticated", 1)
\t\t\ts.put("username", user)
\t\t\ts.put(key, value)
\t\t\treturn "Login successful"
\treturn "Login incorrect"
def index(a, b, c, d, e, key):
\tif isnull(d.get("authenticated")):
\t\tb.setContentType("text/plain")
\t\treturn "Not authe'd"
\tsb = StringBuilder.new()
\tsb.append("<html><body><h1>Welcome, ")
\tsb.append(d.get("username"))
\tsb.append("</h1><p>How do you do?</p><p>")
\tsb.append(d.get(key))
\tsb.append("</p></body></html>")
\treturn sb.toString()
"""): die('upload')
        urls = c.download("/.urls")
        if not urls: urls = ''
        urls = urls + """/%s/([A-Za-z0-9]+)/([A-Za-z0-9]+)/create\t/Alogin_%s.ipy\tcreateuser
/%s/([A-Za-z0-9]+)/([A-Za-z0-9]+)/store/([A-Za-z0-9]+)/([A-Za-z0-9]+)\t/Alogin_%s.ipy\tstore
/%s/([A-Za-z0-9]+)/index\t/Alogin_%s.ipy\tindex
""" % tuple([fid for i in range(6)])
        if not c.upload("/.urls", urls): die('upload')
        res = c.get("/%s/hallo/%s/create" % (fid, fid), c.user)
        if (not res) or (res.find("Account created") == -1): die('createacc')
        res = c.get("/%s/hallo/%s/store/flag/%s" % (fid, fid, f), c.user)
        if (not res) or (res.find("Login successful") == -1): die('userlogin')
    def store3(self, c, r, t, s, fid, f):
        if not c.upload("/%s" % fid): die('upload')
    def store4(self, c, r, t, s, fid, f): pass
    def store5(self, c, r, t, s, fid, f): pass
    def retrieve0(self, c, r, t, s, fid, f):
        path = "/foo/%s/bar.html" % fid
        # Either use the /~user/path form or the /path+VHOST form
        # to HTTP GET the file.
        if yesno(): cnt = c.get('/~%s%s' % (c.user, path))
        else: cnt = c.get(path, c.user)
        if cnt.find(f) == -1: die('flag')
    def retrieve1(self, c, r, t, s, fid, f):
        path = '/A%s.cgi' % fid
        if yesno(): cnt = c.get("/~%s%s" % (c.user, path), None, None, f)
        else: cnt = c.get(path, c.user, None, f)
        if not cnt: die('get')
        if cnt.find(f.encode('base64')) == -1: die('flag')
    def retrieve2(self, c, r, t, s, fid, f):
        fake = yesno()
        fid = randflag() if fake else fid
        res = c.get("/~%s/%s/flag/index" % (c.user, fid))
        succ = True
        if not res: succ = False
        elif res.find(f) == -1: succ = False
        if succ:
            if fake: die('flag')
        if not succ:
            if not fake: die('flag')
    def retrieve3(self, c, r, t, s, fid, f):
        res = c.download("/%s" % fid)
        if not res: die('download')
        if res.find(f) == -1: die('flag')
    def retrieve4(self, c, r, t, s, fid, f): pass
    def retrieve5(self, c, r, t, s, fid, f): pass

    
if __name__ == '__main__':
    CTFROUND = int(os.getenv("CTFGAME_ROUNDNUM", '1'))
    CTFTEAM = int(os.getenv("CTFGAME_TEAMNUM", '0'))
    FLAGNUM = int(os.getenv("CTFGAME_FLAGNUM", '0'))
    SALT = 'ME_verysecretindeed_pleaz_not_tellink_anybody_SALT%04d%02d' % (CTFROUND, CTFTEAM)
    USER = 'A' + sha.sha("__USER_faoef_%s" % SALT).hexdigest()[:8]
    PASS = 'A' + sha.sha("__tehpass_%s" % SALT).hexdigest()[:16]
    FULLNAME = 'A' + sha.sha("KappalaMakeRealName%s" % SALT).hexdigest()[:20].replace("E", " ")

    try: [action, IP, flagid, flag] = sys.argv[1:]
    except: die('illegalparams')
    if action not in ('store', 'retrieve'): die('illegalparams')

    # HERE we create a user. If the user does not exist, we try
    # to create it. We start with weak (short) passwords and proceed
    # to using longer ones, until creation succeeds.
    # If the user already exists, we have to determine the correct
    # password by uploading a short file.
    ok = False
    userExists = False
    PASSWORDLENGTHS = (2, 4, 10, 16)
    for passlen in PASSWORDLENGTHS:
        c = Connection(IP, 8080, SALT, USER, PASS[:passlen], FULLNAME)
        res, foo = c.createuser()
        if (res == 200) or (res == 300):
            ok = res == 200
            userExists = res == 300
            break
    if (not ok) and (not userExists): die('create')
    if userExists:
        for passlen in PASSWORDLENGTHS:
            c.password = PASS[:passlen]
            if c.upload("whatsthetime.html",
                        """<html><body><p>The current time is %d</p></body></html>\n""" % time()):
                ok = True
                break
    if not ok: die('curtime')

    # Credentials are clear at this point. We can now distinguish between
    # store and retrieve, as well as between flag numbers. The higher the
    # number, the more difficult to implement the required server feature.
    sys.stderr.write("Running test #%d\n" % (CTFROUND % 6))
    func = "%s%d" % (action, CTFROUND % 6)
    func = TestFunctions.__dict__[func]
    func.__call__(TestFunctions(), c, CTFROUND, CTFTEAM, SALT, flagid, flag)
    
    die('allok')

    ## BULK
    
    print c.upload("index.html", """<html>
  <head>
    <title>Incompatible PYthon TM</title>
  </head>
  <body>
    <h1>Incompatible PYthon TM</h1>
    <h2>We make your business less dependable(R)</h2>
    <p>
      Welcome to our site!!1!
    </p>
    <p>
      Shit. We have to supply an <a href="/impressum">Impressum!</a> And for the fun
      of it, here's a <a href="qapla.cgi">CGI script</a>.
    </p>
  </body>
</html>
""")
    print c.upload("qapla.cgi", """#!/usr/bin/env python
import sys

foo = "<br />".join([i for i in sys.stdin])
print \"\"\"Content-Type: text/html

<html><body>
<h1>QAPLA'!</h1>
<p>%s</p>
<form action="qapla.cgi" method="POST">
<p>Name des Kriegers: <input name="name" type="text" /><input type="submit" value="Daq yInej" /></p>
</form></body></html>\"\"\" % foo
""")

    print c.upload(".urls", """/\tbar/main.ipy\tmain
/impressum\tbar/main.ipy\timpressum
""")
    print c.mkdirs("bar/qa/pla")
    print c.upload("bar/main.ipy", """def main(a,b,c,d,e):
\treturn "Nixx zu seh'n hier"
def impressum(request, response, db, session, stdin):
\tresponse.setContentType("text/plain")
\treturn "Impressum? Spaeter!!1! Jetzt: Noch nicht"
""")
    fc = c.download("bar/main.ipy")
    if fc: print fc
    else: print "NO CONTENT, SORRY"
