#! /usr/bin/ruby
require 'digest/sha1'
require 'rubygems'
require 'sqlite3'
require 'uri'
require 'net/http' 
srand(Time.now.to_i)

module CTF

  class CTF::Exception < ::Exception
    def initialize(msg = nil)
      super(msg)
    end

    def return_code
      raise "return_code must be overwritten"
    end
  end
  
  class CTF::UnknownState < CTF::Exception
    @result = nil
    def initialize(result,msg = nil)
      super(msg)
      @result = result
    end
    def return_code
       9 
    end
    def result
      @result
    end
  end

  class CTF::ConnectionError < CTF::Exception
    def initialize(msg = nil)
      super(msg)
    end
    def return_code
      1
    end
  end
  
  class WrongFlag < CTF::Exception
    def initialize(msg = nil)
      super(msg)
    end
    def return_code
      5
    end
  end
  
  
  class ServiceLacksFunctionality < CTF::Exception
    def initialize(msg = nil)
      super(msg)
    end
    def return_code
      9 
    end
  end
  
  class SeriveResponseTimeout < CTF::Exception
    def initialize(msg = nil)
      super(msg)
    end
    def return_code
     13 
    end
  end
  
  class StatusUnknown < CTF::Exception
    def initialize(msg = nil)
      super(msg)
    end
    def return_code
     17 
    end
  end
  
  class Error < CTF::Exception
    def initialize(msg = nil)
      super(msg)
    end
    def return_code
      $stderr.print "#{self}\n"
      21
    end
  end
end
@db = SQLite3::Database.new("flags.db")
@db.busy_timeout(4000)
@db.results_as_hash = true

@salt = "124-ask823"
@scheme = "http"
@port = "81"
@main = "cgi-bin/main.rb"
@mode = nil
@ip = nil
@flagID = nil
@flag = nil
@debug  = true
# For normal operation 1 should be used! 
@level = 6 

def store?
  return mode.eql?(:store)
end

def debug(options = {:level => 3,:msg =>""})
  $stderr.print "#{options[:msg]}\r\n" if @debug and @level >= options[:level]
end

def retrieve?
  return mode.eql?(:retrieve)
end

def port
  @port
end

def logout cookie, options = {:timeout => 3} 
  debug :msg => "Logging out.." , :level => 3
  begin
    uri = URI.parse("#{@scheme}://#{@ip}:#{@port}/#{@main}")
  rescue URI::InvalidURIError => e
    debug :msg => "Coult not parse URL in logout. #{e}" , :level => 1
    raise CTF::Error.new("Could not parse the URL in the logout method")
  end
  req = Net::HTTP::Post.new(uri.path)
  req.form_data=({ 'action' => '4'})
  req['Cookie']= cookie
  http = Net::HTTP.new(uri.host,uri.port)
  http.read_timeout= options[:timeout]
  http.open_timeout= options[:timeout]
  debug :msg => "Timeouts (Read: #{http.read_timeout} , Open: #{http.open_timeout})", :level => 6
  res = nil
  begin
    res = http.start {|http| http.request req}
  rescue TimeoutError => e
    debug :msg => "Request takes too much time! Timeout", :level => 2
    raise CTF::CTF::ConnectionError.new("Timeout")
  rescue IOError => e
    debug :msg => "IOError in logout method", :level => 1
    raise CTF::Error.new("IOError raised! Check Testscript")
  end
  result = Hash.new
  if res.kind_of?Net::HTTPSuccess
    debug :msg => "Request was successfull" , :level => 6
    if res.body =~ /Logout successful/m
      debug :msg => "Logout was successfull" , :level => 5
    else
      debug :msg => "Logout failed" , :level =>3 
      debug :msg => "#{res.body}" , :level => 6 
      raise CTF::ServiceLacksFunctionality.new("Logout doesn't worked")
    end
  else
    raise CTF::ServiceLacksFunctionality.new("Request for logout failed: #{res.code} #{res.message}")
  end
end

def createMultipart(flag,form_data)
  debug :msg => "Creating Multipart.." , :level => 5
  str = ""
  boundary = "-----------------------------12384895531882410547944975825"
  content_type = "Content-Type: text/plain"
  str << boundary
  str << "\r\n"
  str << 'Content-Disposition: form-data; name="file"; filename="flag.txt"'
  str << "\r\n"
  str << content_type 
  str << "\r\n\r\n"
  str << flag 
  str << "\r\n"
  form_data.each do |data|
    if data.kind_of?Hash
      str << boundary
      str << "\r\n"
      str << 'Content-Disposition: form-data; name="'+data[:key]+'"'
      str << "\r\n\r\n"
      str << data[:value]
      str << "\r\n"
    end
  end
  str << boundary
  str << "--\r\n"
  b = boundary[2..-1]
  return "multipart/form-data; boundary=#{b}", str
end

def upload_flag(flag, session, options = {:timeout => 20})
  debug :msg => "Uploading flag.." , :level => 3
  begin
    uri = URI.parse("#{@scheme}://#{@ip}:#{@port}/#{@main}")
  rescue URI::InvalidURIError => e
    debug :msg => "Coult not parse URL in upload_flag. #{e}" , :level => 1
    raise CTF::Error.new("Could not parse URL in upload_flag")
  end
  req = Net::HTTP::Post.new(uri.path)
  type, multipart = createMultipart(flag, [:key => "action" , :value => "Upload"])
  req.body=(multipart)
  req['Cookie']= session 
  req.set_content_type type
  http = Net::HTTP.new(uri.host,uri.port)
  http.read_timeout= options[:timeout]
  http.open_timeout= options[:timeout]
  debug :msg => "Timeouts (Read: #{http.read_timeout} , Open: #{http.open_timeout})", :level => 6
  res = nil
  begin
    res = http.start {|http| http.request req}
  rescue TimeoutError => e
    debug :msg => "Request takes too much time! Timeout", :level => 2
    raise CTF::ConnectionError.new("Upload Flag request timed out!")
  rescue IOError => e
    debug :msg => "IOError in upload_flag method. #{e}", :level => 1
    raise CTF::Error.new("IOError in testscript! (upload_flag)")
  end
  link = nil
  if res.kind_of?Net::HTTPSuccess
    debug :msg => "Request was successfull" , :level => 6
    if res.body =~ /File successfully added/m
      begin
        link = /<span id=.dl.>(.*)<\/span>/.match(res.body)[1]
        URI.parse(link)
        return link
      rescue URI::InvalidURIError => e
        debug :msg => "Fetched link is no valid link!", :level => 2 
        raise CTF::Error.new("Could not parse Flag-URL. URL is not valid.")
      rescue Exception => e
        debug :msg => "!", :level => 2 
        raise CTF::Error.new("Could not find URL in Output")
      end
      debug :msg => "File upload successfull. Link to download flog: #{result[:link]}" , :level => 3
    else
      debug :msg => "Upload failed" , :level => 3 
      raise CTF::Error.new("Upload failed. Received unparsable data")
    end
  else
    debug :msg => res.class , :level => 6
    debug :msg => req.to_s , :level => 3 
    raise CTF::Error.new("HTTP-Request to upload a file wasn't successfull: #{res.code}")
  end
end

def change_pass user, oldpass, newpass, newpass2, cookie
 #TODO
end

def create_user user, password, options = {:timeout => 3}
  debug :msg => "Creating a new user account ..." , :level => 3
  uri = nil
  begin
    uri = URI.parse "#{@scheme}://#{ip}:#{port}/#{@main}"
  rescue URI::InvalidURIError => e
    debug :msg => "Coult not parse URL in create_user. #{e}" , :level => 1
    raise CTF::Error.new("Could not parse Flag-URL. URL is not valid.")
  end
  debug :msg => uri , :level => 6
  req = Net::HTTP::Post.new(uri.path)
  req.form_data = { 'action' => 'create', 'username' => user, 'password1' => password, 'password2' => password}
  http = Net::HTTP.new(uri.host,uri.port)
  http.read_timeout= options[:timeout]
  http.open_timeout= options[:timeout]
  debug :msg => "Timeouts (Read: #{http.read_timeout} , Open: #{http.open_timeout})", :level => 6
  res = nil
  begin
    res = http.start do |http|
      http.request(req)
    end
  rescue TimeoutError => e
    debug :msg => "Request takes too much time! Timeout", :level => 2
    raise CTF::ConnectionError.new("Request takes to long")
  rescue IOError => e
    debug :msg => "IOError in create_user method. #{e}", :level => 1
    raise CTF::Error.new("IOError in testscript")
  end
  result = Hash.new
  if res.kind_of?Net::HTTPSuccess
    debug :msg => "Request was successfull" , :level => 6
    cookie =  res['Set-Cookie']
    if cookie.nil?
      raise CTF::Error.new("Response to create a user contains no cookie")
    end
    if /User #{user} was successfully created/m === res.body
      cookie = cookie.split(/;\s?/)
      result[:cookie] = cookie[0]
      result[:user] = user
      result[:password] = password
      debug :msg => "Cookie: #{result[:cookie]}" , :level => 6
      debug :msg => "User #{user} was succuessfully created with password #{password}" , :level => 4
      return result
    else
      result[:cookie] = cookie[0]
      result[:user] = user
      result[:password] = password
      debug :msg => "User #{user} could not created" , :level => 4
      debug :msg => res.body, :level => 6
      raise CTF::UnknownState.new(result,"Not sure if user was created")
    end
  else
      if res.nil?
        raise CTF::Error.new("HTTP-Request failed.")
      else
        raise CTF::Error.new("HTTP-Request failed: #{res.code}")
      end
  end
end

def mode
  return @mode unless @mode.nil?
  unless (ARGV[0] =~ /(STORE|RETRIEVE)/i).nil?
    if ARGV[0].upcase.eql?"STORE"
      @mode = :store 
      return @mode
    elsif ARGV[0].upcase.eql?"RETRIEVE"
      @mode = :retrieve
      return @mode
    else
      usage
      exit
    end
  else
    usage
  end
end

def login user, pass, options = {:timeout => 3}
  debug :msg => "Logging in ..." , :level => 3
  begin
    uri = URI.parse "#{@scheme}://#{ip}:#{port}/#{@main}"
  rescue URI::InvalidURIError => e
    debug :msg => "Coult not parse URL in login. #{e}" , :level => 1
    raise CTF::Error.new("URL to login is not valid")
  end
  debug :msg => uri.to_s , :level => 6
  req = Net::HTTP::Post.new(uri.path)
  req.form_data = { 'action' => 'login', 'username' => user, 'password' => pass}
  http = Net::HTTP.new(uri.host, uri.port)
  http.read_timeout= options[:timeout]
  http.open_timeout= options[:timeout]
  debug :msg => "Timeouts (Read: #{http.read_timeout} , Open: #{http.open_timeout})", :level => 6
  res = nil
  begin
    res = http.start do |http|
      http.request(req)
    end
  rescue TimeoutError => e
    debug :msg => "Request takes too much time! Timeout", :level => 2
    raise CTF::ConnectionError.new("Request takes too long")
  end
  result = Hash.new
  if res.kind_of?Net::HTTPSuccess
    debug :msg => "Request was successfull" , :level => 6
    cookie = res['Set-Cookie']
    unless cookie.nil?
      cookie = cookie.split(/;\s?/)
      if /Login for user #{user} was successful/ === res.body
        cookie = cookie[0]
        debug :msg => "Cookie: #{cookie}" , :level => 6
        debug :msg => "User #{user} was succuessfully logged in with password #{pass}", :level => 4
        return cookie
      end
    end
    raise CTF::Error.new("Login for user failed")
  else
    raise CTF::Error.new("Login for user failed")
  end
end

def ip
  return @ip unless @ip.nil?
  ip = ARGV[1].dup
  if ip =~ /([0-9]{1,3}\.)[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}/
    ip.untaint
    @ip = ip
    return @ip
  else
    usage
    exit
  end
end

def flag
  return @flag unless @flag.nil? 
  @flag = ARGV[3].dup
  if /[0-9a-zA-Z]*/===@flag
    @flag.untaint
  else
    debug :msg => "Flag has an invalid format" , :level => 1
    raise CTF::Error.new("Flag has an invalid format")
  end
end

def flagID
  return @flagID unless @flagID.nil? 
  @flagID = ARGV[2].dup
  if /[0-9a-zA-Z]*/ === @flagID
    @flagID.untaint
  else
    debug :msg => "FlagID has an invalid format" , :level => 1
    raise CTF::Error.new("FlagID has an invalid format")
  end
end

def usage
  puts "Usage"
  puts "#{$0} (STORE|RETRIEVE) IP FLAGID FLAG"
end

def retrieve_flag uri, expected_flag
  debug :msg => "Fetching flag from #{uri} ..." , :level => 3
  result = Hash.new
  unless uri.nil?
    debug :msg => uri , :level => 6
    begin
      url = URI.parse(uri)
    rescue URI::InvalidURIError => e
      debug :msg => "Coult not parse URL in retrieve_flag. #{e}" , :level => 1
      raise CTF::Error.new("Could not parse URL in retrieve_flag") 
    end
    req = Net::HTTP::Get.new(url.path)
    res = Net::HTTP.start(url.host, url.port) {|http|
       http.request(req)
    }
    if res.kind_of?Net::HTTPSuccess
      if res.body.eql?expected_flag
        return res.body
      else
        raise CTF::WrongFlag.new("Received a wrong flag")
      end
    else
      debug :msg => "Flag was not received" , :level => 2
      if res.nil?
        raise CTF::Error.new("HTTP-Response indicates a problem: #{res.code}")
      else
        raise CTF::Error.new("HTTP-Response indicates a problem.")
      end
    end
    return result
  else
    debug :msg => "Uri ist leer", :level => 3
    raise CTF::Error.new("URL to the flag was empty")
  end
end

def add_user flag, ip
  if flag.nil? or ip.nil?
    raise CTF::Error.new("Insufficient arguements to create a user (Problem in testscript)")
  end
  results = nil
  begin
    results = @db.execute("select * from flags where id = ?", flag)
  rescue SQLite3::Exception => e
    debug :msg => "add_user raised an exception: #{e.message}", :level => 2
    raise CTF::Error.new("Error in communication with database. (add_user)")
  end
  user = Hash.new
  if results.length > 0 
    user[:name]  = results.first['user'][0..8]
    user[:pass] = results.first['password'][0..8]
    user[:link] = results.first['link']
    user[:flag] = results.first['id']
    if ip.to_s.eql?ip
      user[:ip] = results.first['ip']
    else
      debug :msg => "FlagID does not match IP, using old one", :level => 2 
      #TODO add error handling
    end
    cookie = login user[:name], user[:pass]
    user[:cookie] = cookie 
    return user
  else
    user[:flag] = flag.to_s
    user[:name] = Digest::SHA1.hexdigest("user#{@salt}#{flag}")[0..8]
    user[:pass] = Digest::SHA1.hexdigest("pass#{@salt}#{flag}")[0..8]
    user[:link] = nil
    user[:ip] = ip.to_s
    result = create_user user[:name], user[:pass]
    begin
      @db.execute("INSERT INTO flags VALUES(?,?,?,?,?)",user[:flag], user[:name], user[:pass],user[:link], user[:ip])
    rescue SQLite3::Exception => e
      raise CTF::Error.new("Error in communication with database. Insert INTO flags (add_user)")
    end
    user[:cookie] = result[:cookie]
    return user
  end
end
def receive_link flag
    begin
      res = @db.execute("SELECT link from flags where id=?", flag)
      if res.length > 0
        return res.first['link']
      else
        raise CTF::Error.new("No link is known to this flag")
      end
    rescue SQLite3::Exception => e
      raise CTF::Error.new("Error in communication with database. Insert INTO flags (add_user)")
    end
end
def store_link flag, link
  unless link.nil?
    debug :msg => "Updating link #{link} for Flag #{flag}", :level => 3 
    begin
      @db.execute("update flags set link = ? where id=?", link, flag) 
    rescue SQLite3::Exception => e
      raise CTF::Error.new("Error in communication with database. Store link")
    end
  end
end

$SAFE=1
if ARGV.length != 4
  usage
  exit
else
  case mode
  when :store
    begin
      debug :msg => "Mode: STORE", :level => 6
      user = nil
      begin
        user=  add_user flag, ip
      rescue CTF::UnknownState => e
        user = e.result
      end
      cookie = user[:cookie]
      link= upload_flag flag, cookie
      store_link flag, link
    rescue CTF::Exception => e
      exit e.return_code
    rescue Exception => e
      debug :msg => e.message, :level => 1
      exit 21
    end
    exit 0 
  when :retrieve
    return_value = 0 
    debug :msg => "Mode: RETRIEVE", :level => 6
    link = nil
    begin
      user=  add_user flag, ip
      result = logout user[:cookie]  
      link = user[:link]
    rescue CTF::Error => e
      debug :msg => "#{e}", :level => 1
      return_value = e.return_code
    rescue CTF::Exception => e
      return_value = e.return_code
    end
    begin
      if link.nil?
        link = receive_link flag
        return_value = 9 
      end
      retrieve_flag(link, flag)
    rescue CTF::Exception => e
      exit e.return_code
    rescue Exception => e
      $stderr.print("#{e.message}\r\n") 
      $stderr.print("#{e.backtrace}\r\n") 
      exit 21
    end
    exit return_value
  else
    usage
    exit
  end
end
