#!/usr/bin/env python2.5

# Testscript template
# (C) 2008, Hans-Christian Esperer
# hc@hcesperer.org
# Licensed under the new (3-clause) BSD license.
# Please adapt to your needs.

from sys import exit, argv
import sys
from os import popen
from time import time, sleep
from random import seed, random, randint
from pprint import pformat
import re
import socket
import hashlib

# Error codes
# Please make sure you have read http://ctf.hcesperer.org/gameserver/errorcodes.html
ERR_OK = 0 # All OK
ERR_CONNECTION = 1 # Connection refused / connection attempt timed out
ERR_WRONGFLAG = 5 # A wrong flag / no flag was returned
ERR_FUNCLACK = 9 # The service lacks functionality
ERR_TIMEOUT = 13 # Done automatically by the scorebot under normal circumstances
ERR_UNKNOWN = 17 # Temporary status -- do not use unless you've got a very good reason to do so
ERR_GENERIC = 21 # Be sure to include a descriptive message if using this error
ERR_PROTOCOL = 25 # Protocol violation

# Add error descriptions here.
# Note (to zeri): do not insult people. ;-)
shit = {'conn':     ("Unable to connect to the service", ERR_CONNECTION),
        'greeting': ("Server didn't greet correctly", ERR_PROTOCOL),
        'smenu':    ('"Subject database" menu item not found', ERR_PROTOCOL),
        'file':     ('"File citizen" menu item not found', ERR_PROTOCOL),
        'filed':    ("Server didn't report our record as filed", ERR_PROTOCOL),
        'back':     ('"BACK" menu item not found', ERR_PROTOCOL),
        'exit':     ('"EXIT" menu item not found', ERR_PROTOCOL),
        'byname':   ('"Search citizens by name" menu item not found', ERR_PROTOCOL),
        'byid':     ('"Search citizens by ID" menu item not found', ERR_PROTOCOL),
        'id':       ("Server didn't report the ID for a record", ERR_PROTOCOL),
        'fname':    ("Server didn't report the first name for a record", ERR_PROTOCOL),
        'lname':    ("Server didn't report the last name for a record", ERR_PROTOCOL),
        'comment':  ("Server didn't report the comment field for a record", ERR_PROTOCOL),
        'wrongname': ("Found the flag, but under the wrong name", ERR_PROTOCOL),
        'wrongname': ("Found the flag, but under the wrong ID", ERR_PROTOCOL),
        'flag':     ("Flag not found", ERR_PROTOCOL),
        'allok':    ("Everything is fine", ERR_OK)}

# output stream aliases
infout = sys.stderr
resout = sys.stdout

# This class can be used to read from a socket linewise. It has built-in
# overflow protection.
class LineReader:
    def __init__(self, socket):
        self.s = socket
        self.buf = ''
    def readline(self):
        while True:
            pos = self.buf.find("\n")
            if pos != -1:
                line, self.buf = self.buf[:pos], self.buf[pos + 1:]
                break
            frag = self.s.recv(8192)
            if len(self.buf) > 8192: raise EOFException()
            frag = frag.replace("\r", "")
            if (len(self.buf) == 0) and (len(frag) == 0): raise EOFException()
            if len(frag) == 0:
                line, self.buf = self.buf, ''
                break # EOF
            self.buf = self.buf + frag
        line = line.replace("\x07", "") # bell chars
        line = line.replace("\x1B[K", "") # line clear sequences
        line = re.sub("\x1B\[[^m]*m", "", line) # color sequences
        return line
    def readlineparams(self):
        line = self.readline()
        # infout.write("RECV: %s" % repr(line))
        line = line.split(" ", 1)
        try: line = line[1]
        except: line = line[0]
        parts = line.split(":", 1)
        sparts = parts[0].strip().split(" ")
        del parts[0]
        return sparts + parts

def die(reason):
    
    """This function should be called to terminate the testscript,
whether successfully or unsuccessfully. An exitcode must always be
specified."""
    try: msg, exitcode = shit[reason]
    except: msg, exitcode = reason, 1
    resout.write(msg)
    resout.flush()
    exit(exitcode)


def randflag():
    """This function can be used to generate random (invalid, but
indistinguishable from real) flags."""
    flagpart = "ABCDEFabcdef0123456789"
    return ''.join([flagpart[randint(0, len(flagpart) - 1)] for i in range(8)])

def expect(lr, pat, err):
    """Wait for a certain pattern to be read on the socket.
The matched pattern is returned if found, otherwise this function
calls die with the specified err."""
    try:
        what = re.compile(pat)
        i = 0
        while True:
            line = lr.readline()
            i = i + 1
            if i > 128: die('toomuch')
            infout.write("%s|  %s\n" % (pat, line))
            res = what.search(line)
            if res != None:
                infout.write('>> found string %s (%s)\n' % (repr(pat), pformat(res.groups())))
                return res.groups()
    except:
        infout.write("Didn't catch string %s\n" % pat)
        die(err)

def select(lr, pat, err):
    """expect and select a specific menu item"""
    item = expect(lr, "([^ ]+) - %s" % pat, err)
    lr.s.send("%s\n" % item)

# getsock can be used to create a TCP socket
def getsock(): return socket.socket(socket.AF_INET, socket.SOCK_STREAM)


def flagname(flag):
    nlast = 88799
    nfirst = 5494

    h = hashlib.sha1()
    h.update(flag)
    h.update("gimmelast")
    ilast = long(h.hexdigest(), 16) % nlast
    h = hashlib.sha1()
    h.update(flag)
    h.update("gimmefirst")
    ifirst = long(h.hexdigest(), 16) % nfirst

    flast = open("names.all.last")
    ffirst = open("names.all.first")

    while ilast > 0:
        flast.readline()
        ilast = ilast - 1
    while ifirst > 0:
        ffirst.readline()
        ifirst = ifirst - 1
    firstname = ffirst.readline().strip()
    lastname = flast.readline().strip()

    flast.close()
    ffirst.close()

    return (firstname, lastname)

def saveid(flag, id):
    f = open("flagids", "a")
    f.write("%s %s\n" % (flag, id))
    f.close()

def readid(flag):
    f = open("flagids")
    res = None
    for line in f:
        (fl, id) = line.split()
        if fl == flag:
            res = id
    return res


# The store function is called to store a flag.
def store((ip, flagid, flag)):
    seed(time())
    s = getsock()

    try: s.connect((ip, 1353))
    except: die('conn')

    # Specify an appropriate timeout
    s.settimeout(10.0)
    # You should always use the LineReader to read from the socket,
    # unless, of course, you implement a custom binary protocol.
    # Use extra care in that case. Use a lot of care, anyway;-)
    lr = LineReader(s)

    (version, station) = expect(lr, 'POL/TER/VOR/AUS (.*)  (.*)  .*', 'greeting')
    infout.write("Station identified itself as %s, service version %s\n" % (station, version))

    if True:
        infout.write(">> trying to file a new citizen\n")
        (firstname, lastname) = flagname(flag)
        select(lr, "Subject", 'smenu')
        select(lr, "File", 'file')
        infout.write(">> first name: %s\n" % firstname)
        s.send(firstname + "\n")
        infout.write(">> last name: %s\n" % lastname)
        s.send(lastname + "\n");
        infout.write(">> saving flag in comment field\n")
        s.send(flag + "\n")
        (id,) = expect(lr, "< Filed as ID ([0-9]+)", 'filed')
        infout.write(">> filed as id %s\n" % id)
        saveid(flag, id)
        select(lr, "BACK", 'back')
    else:
        pass
        
    select(lr, "EXIT", 'exit')
    infout.write("===== SUCCESSFULLY STORED FLAG =====\n")
    die('allok')

# The retrieve function is called to retrieve a flag.
def retrieve((ip, flagid, flag)):
    seed(time())
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(10.0)
    try: s.connect((ip, 1353))
    except: die('conn')
    lr = LineReader(s)

    (version, station) = expect(lr, 'POL/TER/VOR/AUS (.*)  (.*)  .*', 'greeting')
    infout.write("Station identified itself as %s, service version %s\n" % (station, version))

    if random() < 1.0:
        infout.write(">> trying to retrieve citizen record\n")
        id = readid(flag)
        (firstname, lastname) = flagname(flag)
        select(lr, "Subject", 'smenu')
        if random() < 0.5:
            select(lr, "Search for name", 'byname')
            infout.write(">> first name: %s\n" % firstname)
            s.send(firstname + "\n")
            infout.write(">> last name: %s\n" % lastname)
            s.send(lastname + "\n")
        else:
            select(lr, "Search for ID", 'byid')
            infout.write(">> id: %s\n" % id)
            s.send(id + "\n")
        infout.write(">> searching for flag in comment field(s)\n")
        flagfound = False

        (output,) = expect(lr, "(< .*)", 'flag')
        while True:
            # check whether we're done
            if output == "< not found":
                die('flag')
            res = re.match("< ID ([0-9]+)", output)
            if res == None:
                die('id')
            (fid,) = res.groups()
            (fname,) = expect(lr, "< First name.*: (.*)", 'fname')
            (lname,) = expect(lr, "< Last name.*: (.*)", 'lname')
            (comment,) = expect(lr, "< Comment.*: (.*)", 'comment')
            if comment==flag:
                if fname==firstname and lname==lastname:
                    if fid==id:
                        infout.write(">> flag found at ID %s (%s %s)\n"
                                     % (fid, fname, lname));
                        flagfound = True
                    else:
                        die('wrongid')
                else:
                    die('wrongname')
            while True:
                output = lr.readline().strip()
                if len(output)>0: break
            if output[0] != '<': break;
        if not flagfound: die('flag')
        select(lr, "BACK", 'back')
    else:
        pass

    select(lr, "EXIT", 'exit')
    infout.write("===== SUCCESSFULLY RETRIEVED FLAG =====\n")
    die('allok')


def test((ip)):
    resout.write("Testing is integrated in store/retrieve functionality.\n")
    exit(0) # service is fully functional

if __name__ == '__main__':
    {   'store': store,
        'retrieve': retrieve,
        'test': test
    }[argv[1]](tuple(argv[2:]))
try: pass
except SystemExit, e: exit(e.code)
except Exception, e:
    resout.write("ERROR! ")
    resout.write(str(e))
    infout.write("Error: " + str(e))
    infout.write("Usage: %s store|retrieve IP FLAGID FLAG" % argv[0])
    infout.write("       %s test IP" % argv[0])
finally: resout.flush()

#vim:set shiftwidth=4 tabstop=4 expandtab textwidth=79:

